#!/usr/bin/perl

use POSIX;
use IO::Socket;
use IO::Select;
use Socket;
use Fcntl;
use Tie::RefHash;

$debug = 1;
$port = 4567;
$ipfw = "/sbin/ipfw";
$rule_start = 1000;
$pipe_start = 100;

$| = 1;

print "starting manad..\n";


#    
%CLRULE = ();

#     
%USERULEN = ();
%CLUSERULEN = ();

#     
%USEPIPEN = ();
%CLUSEPIPEN = ();

#    
%inbuffer	= ();
%outbuffer	= ();
%ready		= ();

tie %ready, 'Tie::RefHash';

#  
$server = IO::Socket::INET->new( LocalPort => $port, Listen => 10 )
	or die "Can`t make server socket: $@\n";

nonblock( $server );

$SIG{INT} = sub { $server->close(); exit( 0 ); };

$select = IO::Select->new( $server );

$pid = getpid();

open(FILE, ">/var/run/manad.pid");
print FILE $pid;
close(FILE);

#   root   
# chroot( $homedir ) or die "Couldn`t chroot to $homedir: $!\n";

#  :  /,  ,
#    

while( 1 )
{
	my $client;
	my $rv;
	my $data;

	#       

	#   -    ?	
	foreach $client ( $select->can_read( 1 ) )
	{
		if ( $client == $server )
		{
			#   
			$client = $server->accept();
			$select->add( $client );
			nonblock( $client );
		}		
		else
		{
			#  
			$data = '';
			$rv = $client->recv( $data, POSIX::BUFSIZ, 0 );

			unless( defined( $rv ) && length $data )
			{
				#     ,   
				delete $inbuffer{$client};
				delete $outbuffer{$client};
				delete $ready{$client};

				$select->remove( $client );
				close $client;
				next;
			}

			$inbuffer{$client} .= $data;

			# ,         
			#     ,  .   - 
			#  $ready{$client} ,  .
			while( $inbuffer{$client} =~ s/(.*\n)// ) { push( @{$ready{$client}}, $1 ) }
		}
	}
	
	#      ?
	foreach $client ( keys %ready ) { handle( $client ); }

	#  ?
	foreach $client ( $select->can_write( 1 ) )
	{
		#   ,    
		next unless $outbuffer{$client};

		$rv = $client->send( $outbuffer{$client}, 0 );
		unless( defined $rv )
		{
			# ,   
			warn "I was told I could write? but I can`t.\n";
			next;
		}
		if ( $rv == length $outbuffer{$client} || $! == POSIX::EWOULDBLOCK )
		{
			substr( $outbuffer{$client}, 0, $rv ) = '';
			delete $outbuffer{$client} unless length $outbuffer{$client};
		}
		else
		{
			#        - .
			#     .
			delete $inbuffer{$client};
			delete $outbuffer{$client};
			delete $ready{$client};

			$select->remove($client);
			close($client);
			next;
		}
	}
}

# handle( $socket )    
#   $client
sub handle
{
	#    $ready{$client}
	#    $outbuffer{$client}
	my $client = shift;
	my $request;

	foreach $request ( @{$ready{$client}} )
	{
		print "\nrequest=".$request if ( $debug == 1 );

		if ( $request =~ /^test/ )
		{
			my $open_client = "";
			foreach my $kod ( keys %CLRULE )
				{ $open_client .= $open_client eq "" ? $kod : " ".$kod; }
			$outbuffer{$client} .= $open_client."\n";
		}
		elsif ( $request =~ /^add\t([0-9]+)\t(.*)/ )
		{
			my ($kod, $rule) = ($1, $2);
			&delete_rule( $kod ) if ( exists $CLRULE{$kod} );
			&add_rule( $kod, $rule ) if ( !exists $CLRULE{$kod} );
		}
		elsif ( $request =~ /^remove\t([0-9]+)/ )
		{
			&delete_rule( $1 ) if ( exists $CLRULE{$1} );
		}
	}

	delete $ready{$client};
}

# nonblock( $socket )     
sub nonblock
{
	my $socket = shift;
	my $flags;

	$flags = fcntl( $socket, F_GETFL, 0 )
		or die "Can`t get flags for socket: $!\n";
	fcntl( $socket, F_SETFL, $flags | O_NONBLOCK )
		or die "Can`t make socket nonblocking: $!\n";
}

sub add_rule
{
	my $kod = $_[0];
	my $rule = $_[1];
	my %N = ();
	my %P = ();

	$CLRULE{$kod} = $rule;

	while ( $rule =~ /\{N([0-9]+)\}/ )
	{
		my $n = $1;
		my $i = $rule_start - 1;
		my $j = 0;
		while( 1 )
		{
			while( 1 )
			{
				$i++;
				last if ( !exists $USERULEN{$i} );
			}
			$j++;
			last if ( $j == $n );
			last if ( $n == 0 );
		}
		$USERULEN{$i} = $kod;
		$N{$n} = $i;
		$rule =~ s/\{N$n\}/$N{$n}/g;
	}

	while ( $rule =~ /\{P([0-9]+)\}/ )
	{
		my $p = $1;
		my $i = $pipe_start - 1;
		my $j = 0;
		while( 1 )
		{
			while( 1 )
			{
				$i++;
				last if ( !exists $USEPIPEN{$i} );
			}
			$j++;
			last if ( $j == $p );
			last if ( $p == 0 );
		}
		$USEPIPEN{$i} = $kod;
		$P{$p} = $i;
		$rule =~ s/\{P$p\}/$P{$p}/g;
	}

	foreach my $i ( keys %N ) { $CLUSERULEN{$kod} .= exists $CLUSERULEN{$kod} && $CLUSERULEN{$kod} ne "" ? " ".$N{$i} : $N{$i}; }
	foreach my $i ( keys %P ) { $CLUSEPIPEN{$kod} .= exists $CLUSEPIPEN{$kod} && $CLUSEPIPEN{$kod} ne "" ? " ".$P{$i} : $P{$i}; }

	$rule =~ s/\|/; \/sbin\/ipfw -q /g;

	print "$ipfw -q $rule\n" if ( $debug == 1 );
	$err = `$ipfw -q $rule`;
}

sub delete_rule
{
	my $kod = $_[0];

	if ( exists $CLRULE{$kod} )
	{
		my @N = split( / /, $CLUSERULEN{$kod} );

		foreach my $i ( @N )
		{
			print "$ipfw delete $i\n" if ( $debug == 1 );
			$err = `$ipfw delete $i`;
			delete $USERULEN{$i};
		}

		my @P = split( / /, $CLUSEPIPEN{$kod} );

		foreach my $i ( @P )
		{
			print "$ipfw pipe delete $i\n" if ( $debug == 1 );
			$err = `$ipfw pipe delete $i`;
			delete $USEPIPEN{$i};
		}

		delete $CLUSERULEN{$kod};
		delete $CLUSEPIPEN{$kod};
		delete $CLRULE{$kod};
	}
}
